import sys, re, hashlib, base64

py3 = sys.version_info >= (3,0)

arrayTypesMapping = [
  ' ', # VTK_VOID            0
  ' ', # VTK_BIT             1
  'b', # VTK_CHAR            2
  'B', # VTK_UNSIGNED_CHAR   3
  'h', # VTK_SHORT           4
  'H', # VTK_UNSIGNED_SHORT  5
  'i', # VTK_INT             6
  'I', # VTK_UNSIGNED_INT    7
  'l', # VTK_LONG            8
  'L', # VTK_UNSIGNED_LONG   9
  'f', # VTK_FLOAT          10
  'd', # VTK_DOUBLE         11
  'L', # VTK_ID_TYPE        12
]

javascriptMapping = {
    'b': 'Int8Array',
    'B': 'Uint8Array',
    'h': 'Int16Array',
    'H': 'Int16Array',
    'i': 'Int32Array',
    'I': 'Uint32Array',
    'l': 'Int32Array',
    'L': 'Uint32Array',
    'f': 'Float32Array',
    'd': 'Float64Array'
}

if py3:
    def iteritems(d, **kwargs):
        return iter(d.items(**kwargs))
else:
    def iteritems(d, **kwargs):
        return d.iteritems(**kwargs)

if sys.version_info >= (2,7):
    buffer = memoryview
    base64Encode = lambda x: base64.b64encode(x).decode('utf-8')
else:
    buffer = buffer
    base64Encode = lambda x: x.encode('base64')

def hashDataArray(dataArray):
    hashedBit = base64Encode(hashlib.md5(buffer(dataArray)).digest()).strip()
    md5sum = re.sub('==$', '', hashedBit)
    typeCode = arrayTypesMapping[dataArray.GetDataType()]
    return '%s_%d%s' % (md5sum, dataArray.GetSize(), typeCode)

def getJSArrayType(dataArray):
    return javascriptMapping[arrayTypesMapping[dataArray.GetDataType()]]



"""
A collection of utilities for cinema camera handling.
"""

import numpy as np
import math


def convert_pose_to_camera(iPosition, iFocalPoint, iViewUp, pose, camType):
    """Converts input eye, focal point and view up using the pose and camType
    and returns (neweye, newfp, newviewup)"""
    def VecNormalize(V1):
        return V1/np.linalg.norm(V1)

    def VecMatrixMul(V, M):
        return [
            V[0]*M[0][0] + V[1]*M[0][1] + V[2]*M[0][2],
            V[0]*M[1][0] + V[1]*M[1][1] + V[2]*M[1][2],
            V[0]*M[2][0] + V[1]*M[2][1] + V[2]*M[2][2]
            ]

    def MatrixMatrixMul(mtx_a, mtx_b):
        tpos_b = list(zip(*mtx_b))
        rtn = [
            [sum(ea*eb for ea, eb in zip(a, b)) for b in tpos_b]
            for a in mtx_a]
        return rtn

    iViewUp = VecNormalize(iViewUp)

    atvec = VecNormalize(np.subtract(iPosition, iFocalPoint))
    rightvec = np.cross(atvec, iViewUp)
    m = [[rightvec[0], rightvec[1], rightvec[2]],
         [iViewUp[0], iViewUp[1], iViewUp[2]],
         [atvec[0],  atvec[1],  atvec[2]]]
    # OK because orthogonal thus invert equals trans
    mInv = [[m[0][0], m[1][0], m[2][0]],
            [m[0][1], m[1][1], m[2][1]],
            [m[0][2], m[1][2], m[2][2]]]
    mi = MatrixMatrixMul(mInv, pose)
    mf = MatrixMatrixMul(mi, m)

    newUp = VecMatrixMul(iViewUp, mf)

    if camType == "azimuth-elevation-roll":
        pi = np.subtract(iPosition, iFocalPoint)
        pr = VecMatrixMul(pi, mf)
        pf = np.add(pr, iFocalPoint)
        return (pf, iFocalPoint, newUp)
    elif camType == "yaw-pitch-roll":
        pi = np.subtract(iFocalPoint, iPosition)
        pr = VecMatrixMul(pi, mf)
        pf = np.add(pr, iPosition)
        return (iPosition, pf, newUp)
    else:
        print("ERROR unexpected camera type")
        return (iPosition, iFocalPoint, iViewUp)


def nearest_camera(poses, mnext):
    """
    find index of the pose that is closest match to mnext
    """
    # TODO: profile, and then if important, optimize this
    # 1: only need to compare 2 of 3 vectors since they are orthogonal
    # 2: will be faster still if we push the matrices into numpy
    # (flatten too perhaps)

    # algorithm: reduce to distance value, and then pick the index of
    # pose with least distance to mnext
    best = -1
    bestdist = -1
    primary = 0
    secondary = 1
    tertiary = 2
    for i in range(0, len(poses)):
        cand = poses[i]
        dist1 = math.sqrt(
            math.pow(mnext[primary][0]-cand[primary][0], 2) +
            math.pow(mnext[primary][1]-cand[primary][1], 2) +
            math.pow(mnext[primary][2]-cand[primary][2], 2))
        dist2 = math.sqrt(
            math.pow(mnext[secondary][0]-cand[secondary][0], 2) +
            math.pow(mnext[secondary][1]-cand[secondary][1], 2) +
            math.pow(mnext[secondary][2]-cand[secondary][2], 2))
        dist3 = math.sqrt(
            math.pow(mnext[tertiary][0]-cand[tertiary][0], 2) +
            math.pow(mnext[tertiary][1]-cand[tertiary][1], 2) +
            math.pow(mnext[tertiary][2]-cand[tertiary][2], 2))
        dist = dist1*dist1+dist2*dist2+dist3*dist3
        if best == -1:
            bestdist = dist
            best = i
        if dist < bestdist:
            # print "NB", dist, i, dist-bestdist
            bestdist = dist
            best = i
    return best



print("import: 'paraview'")
import paraview

print("import: 'paraview.collaboration'")
import paraview.collaboration

print("import: 'paraview.coprocessing'")
import paraview.coprocessing

print("import: 'paraview.cpexport'")
import paraview.cpexport

print("import: 'paraview.cpstate'")
import paraview.cpstate

print("import: 'paraview.lookuptable'")
import paraview.lookuptable

print("import: 'paraview.numeric'")
import paraview.numeric

print("import: 'paraview.numpy_support'")
import paraview.numpy_support

print("import: 'paraview.python_view'")
import paraview.python_view

print("import: 'paraview.cpstate'")
import paraview.cpstate

print("import: 'paraview.servermanager'")
import paraview.servermanager

print("import: 'paraview.simple'")
import paraview.simple

print("import: 'paraview.util'")
import paraview.util




r"""collaboration is a helper module useful for handling synchronisation
in multi-client configuration as well as providing other collaboration
related methods

A simple example:
  from paraview import collaboration

  collaboration.processServerEvents()
"""
#==============================================================================
#
#  Program:   ParaView
#  Module:    collaboration.py
#
#  Copyright (c) Kitware, Inc.
#  All rights reserved.
#  See Copyright.txt or http://www.paraview.org/HTML/Copyright.html for details.
#
#     This software is distributed WITHOUT ANY WARRANTY; without even
#     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#     PURPOSE.  See the above copyright notice for more information.
#
#==============================================================================

import paraview, re, types, sys, vtk
from paraview import servermanager

def processServerEvents():
  """Update the local state based on the notifications received from the server
  that have been generated by other clients."""
  if servermanager.ActiveConnection:
     session = servermanager.ActiveConnection.Session
     if session.IsMultiClients() and session.IsNotBusy():
        while vtkPVClientServerCoreCore.vtkProcessModule.GetProcessModule().GetNetworkAccessManager().ProcessEvents(100):
          pass



"""
Module for handling color lookup tables.
"""

import json
import numpy as np
import math


class LookupTable:
    '''
    Color table container.
    - self.lut          : Actual color LUT.
    - self.x            : Vector containing the value bins.
    - self.adjustedBins : Vector containing the value bins
    Where adjustedBins is adjusted to hold a value > 1.0 in the end.
    '''
    def __init__(self):
        self.name = 'None'
        self.colorSpace = 'None'
        self.lut = None
        self.x = None
        self.adjustedBins = None

    def ingest(self, rgbPoints):
        xs = []
        tlut = []
        minx = float("inf")
        maxx = float("-inf")
        for i in xrange(0, len(rgbPoints), 4):
            x = rgbPoints[i]
            maxx = x if x > maxx else maxx
            minx = x if x < minx else minx
            xs.append(rgbPoints[i])
            rgb = (rgbPoints[i + 1],
                   rgbPoints[i + 2],
                   rgbPoints[i + 3])
            tlut.append(rgb)

        # Remap xs from [minx, max] -> [0, 1]
        scale = maxx - minx
        xs = [(xn - minx) / scale for xn in xs]

        for i in range(0, len(tlut)):
            tlut[i] = (tlut[i][0]*255, tlut[i][1]*255, tlut[i][2]*255)
        lut = np.array(tlut, dtype=np.uint8)
        self.lut = lut
        self.x = xs

        # use a histogram to support non-uniform colormaps (fetch bin indices)
        colorLutXValue = self.x
        bins = list(colorLutXValue)
        if colorLutXValue[-1] < 1.0:
            bins.append(1.0)
        else:
            bins.append(1.01)  # 1.0 gets its own color so need an extra bin
        self.adjustedBins = bins


class LookupTableManager:
    '''
    Loads a set of color lookup tables from a .json file.
    '''
    def __init__(self):
        self.luts = []
        self.lut = None
        self.x = None

        # Add a default color map ('None')
        self.luts.append(LookupTable())

    def read(self, file_path):
        with open(file_path) as json_file:
            data = json.load(json_file)
            for i in range(0, len(data)):
                lutentry = LookupTable()
                lutentry.name = data[i]['Name']
                lutentry.colorSpace = data[i]['ColorSpace']
                lutentry.ingest(data[i]['RGBPoints'])
                self.luts.append(lutentry)

    def getColorLutStructByName(self, name):
        for lut in self.luts:
            if lut.name == name:
                return lut
        return None

    def names(self):
        """
        Return an array of the names of all available lookup tables.
        """
        names = []
        for lut in self.luts:
            names.append(lut.name)
        return names


def add_spectral(luts):
    lutentry = {}
    lutentry['ColorSpace'] = 'RGB'
    lutentry['Name'] = 'Spectral'
    lutentry['NanColor'] = [
        0.6196078431372549, 0.00392156862745098, 0.2588235294117647]
    tlut = [
        0.0, 0.6196078431372549, 0.00392156862745098, 0.2588235294117647,
        0.1, 0.8352941176470589, 0.2431372549019608, 0.3098039215686275,
        0.2, 0.9568627450980393, 0.4274509803921568, 0.2627450980392157,
        0.3, 0.9921568627450981, 0.6823529411764706, 0.3803921568627451,
        0.4, 0.996078431372549, 0.8784313725490196, 0.5450980392156862,
        0.5, 1, 1, 0.7490196078431373,
        0.6, 0.9019607843137255, 0.9607843137254902, 0.596078431372549,
        0.7, 0.6705882352941176, 0.8666666666666667, 0.6431372549019608,
        0.8, 0.4, 0.7607843137254902, 0.6470588235294118,
        0.9, 0.196078431372549, 0.5333333333333333, 0.7411764705882353,
        1.0, 0.3686274509803922, 0.3098039215686275, 0.635294117647058
        ]
    lutentry['RGBPoints'] = tlut
    luts.append(lutentry)


def add_grayscale(luts):
    lutentry = {}
    lutentry['ColorSpace'] = 'RGB'
    lutentry['Name'] = "Grayscale"
    lutentry['NanColor'] = [0, 0, 0]
    s = 32.0
    tlut = [x / s for x in range(0, int(s)+1) for i in range(0, 4)]
    lutentry['RGBPoints'] = tlut
    luts.append(lutentry)


def add_rainbow(luts):
    lutentry = {}
    lutentry['ColorSpace'] = 'RGB'
    lutentry['Name'] = "Rainbow"
    lutentry['NanColor'] = [0, 0, 0]
    tlut = []
    for x in range(0, 64):
        r = 0
        g = 0
        b = 0
        a = (1.0-x/63.0)/0.25
        X = math.floor(a)
        Y = a-X
        if X == 0:
            r = 1.0
            g = Y
            b = 0
        if X == 1:
            r = 1.0-Y
            g = 1.0
            b = 0
        if X == 2:
            r = 0
            g = 1.0
            b = Y
        if X == 3:
            r = 0
            g = 1.0-Y
            b = 1.0
        if X == 4:
            r = 0
            g = 0
            b = 1.0
        tlut.append(x / 63.0)
        tlut.append(r)
        tlut.append(g)
        tlut.append(b)
    lutentry['RGBPoints'] = tlut
    luts.append(lutentry)


def add_ocean(luts):
    lutentry = {}
    lutentry['ColorSpace'] = 'RGB'
    lutentry['Name'] = 'Ocean'
    lutentry['NanColor'] = [0, 0, 0]
    tlut = [
        0.0, 0.039215, 0.090195, 0.25098,
        0.125, 0.133333, 0.364706, 0.521569,
        0.25, 0.321569, 0.760784, 0.8,
        0.375, 0.690196, 0.960784, 0.894118,
        0.5, 0.552941, 0.921569, 0.552941,
        0.625, 0.329412, 0.6, 0.239216,
        0.75, 0.211765, 0.349020, 0.078435,
        0.875, 0.011765, 0.207843, 0.023525,
        1.0, 0.286275, 0.294118, 0.30196
        ]
    lutentry['RGBPoints'] = tlut
    luts.append(lutentry)

if __name__ == "__main__":
    luts = []
    add_spectral(luts)
    add_grayscale(luts)
    add_rainbow(luts)
    add_ocean(luts)

    formatted = json.dumps(
        luts, sort_keys=True, indent=2,
        separators=(',', ': '))
    with open('builtin_tables.json', 'w') as output:
        output.write(formatted)
